import copy
import json
import logging
import os.path
import tempfile
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from mlflow.entities import ArtifactType, CustomMetric, Model, ModelSchema
from mlflow.tracking import MlflowClient
from pydantic import BaseModel, Extra

from mlfoundry.artifact.truefoundry_artifact_repo import MlFoundryArtifactsRepository
from mlfoundry.enums import ModelFramework
from mlfoundry.exceptions import MlFoundryException
from mlfoundry.log_types.artifacts.constants import (
    FILES_DIR,
    INTERNAL_METADATA_PATH,
    MODEL_DIR_NAME,
    MODEL_SCHEMA_UPDATE_FAILURE_HELP,
)
from mlfoundry.log_types.artifacts.utils import (
    _copy_additional_files,
    _get_mlflow_client,
    _validate_artifact_metadata,
    _validate_description,
)

logger = logging.getLogger("mlfoundry")


# TODO: Add some progress indicators for upload and download
# TODO: Support async download and upload


class ModelVersionInternalMetadata(BaseModel):
    class Config:
        extra = Extra.allow

    files_dir: str  # relative to root
    model_dir: str  # relative to `files_dir`
    model_is_null: bool = False
    framework: ModelFramework = ModelFramework.UNKNOWN

    def dict(self, *args, **kwargs):
        dct = super().dict(*args, **kwargs)
        dct["framework"] = dct["framework"].value
        return dct


class ModelVersionDownloadInfo(BaseModel):
    download_dir: str
    model_dir: str
    model_framework: ModelFramework = ModelFramework.UNKNOWN


class ModelVersion:
    def __init__(self, fqn: str) -> None:
        self._mlflow_client = _get_mlflow_client()
        self._model_version = self._mlflow_client.get_model_version_by_fqn(fqn=fqn)
        self._model: Model = self._mlflow_client.get_model_by_id(
            model_id=self._model_version.model_id
        )
        self._mlfa_repo = MlFoundryArtifactsRepository(
            version_id=self._model_version.id, mlflow_client=self._mlflow_client
        )
        self._deleted = False
        self._description: str = ""
        self._metadata: Dict[str, Any] = {}
        self._model_schema: Optional[ModelSchema] = None
        self._set_mutable_attrs()

    def _ensure_not_deleted(self):
        if self._deleted:
            raise MlFoundryException(
                "Model Version was deleted, cannot perform updates on a deleted version"
            )

    def _set_mutable_attrs(self, refetch=False):
        if refetch:
            self._model_version = self._mlflow_client.get_model_version_by_id(
                version_id=self._model_version.id
            )
        self._description = self._model_version.description or ""
        self._metadata = copy.deepcopy(self._model_version.artifact_metadata)
        self._model_schema = copy.deepcopy(self._model_version.model_schema)

    def __repr__(self):
        return f"{self.__class__.__name__}(fqn={self.fqn!r}, model_schema={self._model_schema!r})"

    @property
    def name(self) -> str:
        return self._model.name

    @property
    def model_fqn(self) -> str:
        return self._model.fqn

    @property
    def version(self) -> int:
        return self._model_version.version

    @property
    def fqn(self) -> str:
        return self._model_version.fqn

    @property
    def step(self) -> Optional[int]:
        return self._model_version.step

    @property
    def description(self) -> Optional[str]:
        return self._description

    @description.setter
    def description(self, value: str):
        _validate_description(value)
        self._description = value

    @property
    def metadata(self) -> Dict[str, Any]:
        return self._metadata

    @metadata.setter
    def metadata(self, value: Dict[str, Any]):
        _validate_artifact_metadata(value)
        self._metadata = value

    @property
    def model_schema(self) -> Optional[ModelSchema]:
        return self._model_schema

    @model_schema.setter
    def model_schema(self, value: Union[Dict[str, Any], ModelSchema]):
        if not isinstance(value, ModelSchema):
            value = ModelSchema.parse_obj(value)
        self._model_schema = value

    def raw_download(self, path: Optional[Union[str, Path]]) -> str:
        logger.info("Downloading model version contents, this might take a while ...")
        return self._mlfa_repo.download_artifacts(artifact_path="", dst_path=path)

    def _download(
        self, path: Optional[Union[str, Path]]
    ) -> Tuple[ModelVersionInternalMetadata, ModelVersionDownloadInfo]:
        self._ensure_not_deleted()
        download_dir = self.raw_download(path=path)
        internal_metadata_path = os.path.join(download_dir, INTERNAL_METADATA_PATH)
        if not os.path.exists(internal_metadata_path):
            raise MlFoundryException(
                f"Model version seems to be corrupted or in invalid format due to missing model metadata. "
                f"You can still use .raw_download(path='/your/path/here') to download and inspect files."
            )
        with open(internal_metadata_path) as f:
            internal_metadata = ModelVersionInternalMetadata.parse_obj(json.load(f))
        download_info = ModelVersionDownloadInfo(
            download_dir=os.path.join(download_dir, internal_metadata.files_dir),
            model_dir=os.path.join(
                download_dir, internal_metadata.files_dir, internal_metadata.model_dir
            ),
            model_framework=internal_metadata.framework,
        )
        return internal_metadata, download_info

    def download(self, path: Optional[Union[str, Path]]) -> ModelVersionDownloadInfo:
        _, download_info = self._download(path=path)
        return download_info

    def load(self, **load_model_kwargs):
        from mlfoundry.frameworks import get_model_registry

        internal_metadata, download_info = self._download(path=None)
        if internal_metadata.model_is_null:
            raise MlFoundryException(
                f"Cannot load model as no model object was provided during logging."
            )
        if internal_metadata.framework == ModelFramework.UNKNOWN:
            raise MlFoundryException(
                "Cannot deserialize model as model framework is unknown. You can still download the "
                "contents by calling `.download(path='your/path/here') and load the model manually."
            )
        return get_model_registry(internal_metadata.framework).load_model(
            download_info.model_dir, **load_model_kwargs
        )

    def delete(self) -> bool:
        self._ensure_not_deleted()
        self._mlflow_client.delete_artifact_version(version_id=self._model_version.id)
        self._deleted = True
        return True

    def update(self):
        self._ensure_not_deleted()
        kwargs = {}
        if self.model_schema is not None:
            kwargs["model_schema"] = self.model_schema
        self._model_version = self._mlflow_client.update_model_version(
            version_id=self._model_version.id,
            description=self.description,
            artifact_metadata=self.metadata,
            **kwargs,
        )
        self._set_mutable_attrs()


def _log_model_version(
    run,
    name: str,
    model: Any,
    framework: ModelFramework,
    model_save_kwargs: Optional[Dict[str, Any]] = None,
    additional_files: Sequence[Tuple[Union[str, Path], Optional[str]]] = (),
    description: Optional[str] = None,
    metadata: Optional[Dict[str, Any]] = None,
    model_schema: Optional[Union[Dict[str, Any], ModelSchema]] = None,
    custom_metrics: Optional[Union[List[Dict[str, Any]], CustomMetric]] = None,
    step: Optional[int] = None,
) -> ModelVersion:

    from mlfoundry.frameworks import get_model_registry

    custom_metrics = custom_metrics or []
    model_save_kwargs = model_save_kwargs or {}
    metadata = metadata or {}
    additional_files = additional_files or {}

    # validations
    if framework is None:
        framework = ModelFramework.UNKNOWN
    elif not isinstance(framework, ModelFramework):
        framework = ModelFramework(framework)

    if model is not None and framework == ModelFramework.UNKNOWN:
        raise MlFoundryException(
            "`framework` cannot be passed as 'unknown' when `model` is not None"
        )

    _validate_description(description)
    _validate_artifact_metadata(metadata)

    if model_schema is not None and not isinstance(model_schema, ModelSchema):
        model_schema = ModelSchema.parse_obj(model_schema)

    if custom_metrics and not model_schema:
        raise MlFoundryException(
            "Custom Metrics defined without adding the Model Schema"
        )
    custom_metrics = [
        CustomMetric.parse_obj(cm) if not isinstance(cm, CustomMetric) else cm
        for cm in custom_metrics
    ]

    logger.info("Logging model and additional files, this might take a while ...")
    mlflow_client: MlflowClient = run.mlflow_client
    temp_dir = tempfile.TemporaryDirectory(prefix="truefoundry-")

    internal_metadata = ModelVersionInternalMetadata(
        framework=framework,
        files_dir=FILES_DIR,
        model_dir=MODEL_DIR_NAME,
        model_is_null=model is None,
    )

    try:
        local_files_dir = os.path.join(temp_dir.name, internal_metadata.files_dir)
        os.makedirs(local_files_dir, exist_ok=True)

        # serialize model
        logger.info("Serializing model files to model version contents")
        local_model_dir = os.path.join(local_files_dir, internal_metadata.model_dir)
        if model is not None:
            model_registy = get_model_registry(framework)
            model_registy.save_model(
                model=model, path=local_model_dir, **model_save_kwargs
            )
        os.makedirs(
            local_model_dir, exist_ok=True
        )  # in case model was None, we still create an empty dir

        # verify additional files and paths, copy additional files
        if additional_files:
            logger.info("Adding `additional_files` to model version contents")
            _copy_additional_files(
                root_dir=temp_dir.name,
                files_dir=internal_metadata.files_dir,
                model_dir=internal_metadata.model_dir,
                additional_files=additional_files,
            )
    except Exception as e:
        temp_dir.cleanup()
        raise MlFoundryException("Failed to log model") from e

    # save internal metadata
    local_internal_metadata_path = os.path.join(temp_dir.name, INTERNAL_METADATA_PATH)
    os.makedirs(os.path.dirname(local_internal_metadata_path), exist_ok=True)
    with open(local_internal_metadata_path, "w") as f:
        json.dump(internal_metadata.dict(), f)

    # create entry
    version_id = mlflow_client.create_artifact_version(
        experiment_id=int(run._experiment_id),
        artifact_type=ArtifactType.MODEL,
        name=name,
    )
    artifacts_repo = MlFoundryArtifactsRepository(
        version_id=version_id, mlflow_client=mlflow_client
    )
    try:
        logger.info("Packaging and uploading files to remote ...")
        artifacts_repo.log_artifacts(local_dir=temp_dir.name, artifact_path=None)
    except Exception as e:
        mlflow_client.notify_failure_for_artifact_version(version_id=version_id)
        raise MlFoundryException("Failed to log model") from e
    finally:
        temp_dir.cleanup()
    mlflow_client.finalize_artifact_version(version_id=version_id, run_uuid=run.run_id)
    model_version = mlflow_client.create_model_version(
        artifact_version_id=version_id,
        description=description,
        artifact_metadata=metadata,
        data_path=INTERNAL_METADATA_PATH,
        step=step,
    )
    # update model schema at end
    if model_schema is not None:
        try:
            model_version = mlflow_client.update_model_version(
                version_id=version_id,
                model_schema=model_schema,
            )
            model_version = mlflow_client.add_custom_metrics_to_model_version(
                version_id=version_id, custom_metrics=custom_metrics
            )
        except Exception:
            # TODO (chiragjn): what is the best exception to catch here?
            logger.error(MODEL_SCHEMA_UPDATE_FAILURE_HELP.format(fqn=model_version.fqn))

    return ModelVersion(fqn=model_version.fqn)



import os
import posixpath
import re
import tempfile
from contextlib import contextmanager

from mlfoundry.exceptions import MlFoundryException

KEY_REGEX = re.compile(r"^[a-zA-Z0-9-_]+$")


def validate_key_name(key: str):
    if not key or not KEY_REGEX.match(key):
        raise MlFoundryException(
            f"Invalid run image key: {key} should only contain alphanumeric, hyphen or underscore"
        )


@contextmanager
def save_artifact_helper(run: "mlfoundry.MlFoundryRun", artifact_path: str):
    if not artifact_path:
        raise MlFoundryException("artifact_path cannot be None or empty string")
    artifact_path = posixpath.normpath(artifact_path)
    file_name = posixpath.basename(artifact_path)
    artifact_dir = posixpath.dirname(artifact_path)

    # Setting to None to make sure the file gets saved at the
    # run's root artifact_path if the artifact_path is like "foo.txt"
    artifact_dir = artifact_dir if artifact_dir else None
    if not file_name:
        raise MlFoundryException(f"could not infer file name from {artifact_path}")

    with tempfile.TemporaryDirectory() as local_dir:
        local_path = os.path.join(local_dir, file_name)
        yield local_path
        run.mlflow_client.log_artifact(
            run_id=run.run_id, local_path=local_path, artifact_path=artifact_dir
        )



import enum
import typing

import pandas as pd
from pydantic import BaseModel
from whylogs import DatasetProfile

from mlfoundry.dataset.whylogs_types import DataFrameSummary
from mlfoundry.log_types import DatasetSchemaRunLog, DatasetStatsRunLog, StatsSource


class FieldType(enum.Enum):
    NUMERICAL = "NUMERICAL"
    CATEGORICAL = "CATEGORICAL"
    EMPTY = "EMPTY"


class InferredFieldType(BaseModel):
    field_type: typing.Optional[FieldType]
    pandas_inferred_dtype: str


class Profiles(BaseModel):
    features: DatasetProfile
    actuals: typing.Optional[DatasetProfile] = None
    predictions: typing.Optional[DatasetProfile] = None

    class Config:
        arbitrary_types_allowed = True


class DataSet(BaseModel):
    dataset_name: str

    features: pd.DataFrame
    actuals: typing.Optional[pd.Series] = None
    predictions: typing.Optional[pd.Series] = None

    class Config:
        arbitrary_types_allowed = True
        allow_mutation = False


class DatasetStats(BaseModel):
    features: DataFrameSummary
    actuals: typing.Optional[DataFrameSummary] = None
    predictions: typing.Optional[DataFrameSummary] = None

    def to_log(self) -> DatasetStatsRunLog:
        return DatasetStatsRunLog(value=self.dict(), stats_source=StatsSource.WHYLOGS)


class DatasetSchema(BaseModel):
    features: typing.Dict[str, InferredFieldType]

    # only supports scalar output at this point
    # we do not support multi-label
    actuals: typing.Optional[InferredFieldType] = None
    predictions: typing.Optional[InferredFieldType] = None

    def to_log(self) -> DatasetSchemaRunLog:
        return DatasetSchemaRunLog(value=self.dict())



import logging
import queue
from timeit import default_timer
from typing import Any, List, Optional

from mlflow.entities import Metric
from mlflow.exceptions import MlflowException
from mlflow.tracking import MlflowClient

from mlfoundry import MlFoundryException
from mlfoundry.background import utils as bg_utils
from mlfoundry.background.events import LogMetricsEvent
from mlfoundry.background.interface import Interface

logger = logging.getLogger(__name__)


class SenderJob(bg_utils.BackgroundJob):
    name_prefix = "MLFSenderThread"

    def __init__(
        self,
        interface: Interface,
        mlflow_client: MlflowClient,
        interval: float,
        max_wait_time_for_events: float = 2,
    ):
        """
        Args:
            interface: `background.interface.Interface` instance to use to get and log events with
            interval: time in seconds to sleep between calls to `self._loop` i.e. between iterations
        """
        super().__init__(
            name=f"{SenderJob.name_prefix}:{interface.run_id}", interval=interval
        )
        self._mlflow_client = mlflow_client
        self._interface = interface
        self._max_wait_time_for_events = max_wait_time_for_events

    def _log_metrics(self, metrics: List[Metric]) -> None:
        """
        Log metrics using the MlFlowClient instance

        Args:
            metrics: list of mlflow.entities.metric.Metric instances

        Raises:
            MlFoundryException: in case .log_batch call fails
        """
        try:
            self._mlflow_client.log_batch(
                run_id=self._interface.run_id, metrics=metrics
            )
        except MlflowException as e:
            raise MlFoundryException(e.message).with_traceback(
                e.__traceback__
            ) from None

    def join(self, timeout: Optional[int] = None) -> None:
        """
        Wait for event queue to be empty and then join on self (thread)
        Note: This should not be called from the Sender Thread itself otherwise will cause a deadlock

        Args:
            timeout: time in seconds to wait when joining
        """
        start = default_timer()
        self._interface.event_queue.join()
        time_elapsed = default_timer() - start
        time_left = None if timeout is None else max(0.0, timeout - time_elapsed)
        super().join(timeout=time_left)

    def _consume_event(self) -> Any:
        """
        Consume a single event from the event queue and mark it done

        Returns:
            Any: ...

        Raises:
            queue.Empty: in case the event queue has no events
        """
        # TODO (chiragjn): implement batch consume
        event = self._interface.get_event(
            block=True, timeout=self._max_wait_time_for_events
        )

        try:
            if isinstance(event, LogMetricsEvent):
                # TODO (chiragjn): Think of a better design to publish and consume events
                return self._log_metrics(metrics=event.metrics)
            else:
                logger.warning(
                    bg_utils.make_log(
                        f"Unknown Event of type {type(event)!r} emitted. Dropping event"
                    )
                )
        except Exception as e:
            logger.warning(
                bg_utils.make_log(
                    f"Encountered an internal error while trying to consume: {str(e)}"
                )
            )
        finally:
            self._interface.event_queue.task_done()

    def _loop(self) -> None:
        """
        Attempt to consume a single event from the event queue
        """
        try:
            self._consume_event()
        except queue.Empty:
            pass

    def _finish(self) -> None:
        """
        Consume remaining events in the event queue
        """
        while not self._interface.event_queue.empty():
            try:
                self._consume_event()
            except queue.Empty:
                break
        self._interface = None



import atexit
import json
import logging
import os
import queue
import threading
import typing
import uuid
from datetime import datetime

from mlflow.tracking import MlflowClient

from mlfoundry.exceptions import MlFoundryException
from mlfoundry.monitoring.entities import (
    Actual,
    ActualPacket,
    BasePacket,
    Prediction,
    PredictionPacket,
)
from mlfoundry.monitoring.store.constants import (
    FLUSH_BATCH_SIZE_ENV_VAR,
    FLUSH_INTERVAL_ENV_VAR,
    MAX_QUEUE_SIZE_ENV_VAR,
)
from mlfoundry.monitoring.store.repositories import get_monitoring_store
from mlfoundry.monitoring.store.worker import MonitoringStoreWorker

logger = logging.getLogger(__name__)


def _check_shutdown(func):
    def wrapper(self, *args, **kwargs):
        if self._is_shutdown:
            raise MlFoundryException("cannot execute after client is shutdown")
        return func(self, *args, **kwargs)

    return wrapper


class MonitoringClient:
    def __init__(self, monitoring_uri: str):
        self._is_shutdown = False

        self.monitoring_store = get_monitoring_store(host_uri=monitoring_uri)
        max_queue_size = int(os.getenv(MAX_QUEUE_SIZE_ENV_VAR, "10000"))
        flush_interval = float(os.getenv(FLUSH_INTERVAL_ENV_VAR, "5"))
        flush_batch_size = int(os.getenv(FLUSH_BATCH_SIZE_ENV_VAR, "300"))

        if max_queue_size <= 0:
            raise MlFoundryException(
                f"{MAX_QUEUE_SIZE_ENV_VAR} should be a positive number"
            )

        self.task_queue: "queue.Queue[BasePacket]" = queue.Queue(max_queue_size)
        self.worker_terminate_event = threading.Event()
        self.mlflow_client = MlflowClient()

        if flush_interval <= 0:
            raise MlFoundryException(
                f"{FLUSH_INTERVAL_ENV_VAR} should be a positive number"
            )
        if flush_batch_size <= 0 or flush_batch_size > max_queue_size:
            raise MlFoundryException(
                f"{FLUSH_BATCH_SIZE_ENV_VAR} should be positive"
                f" and lower than {MAX_QUEUE_SIZE_ENV_VAR}"
            )

        self.worker: MonitoringStoreWorker = MonitoringStoreWorker(
            task_queue=self.task_queue,
            terminate_event=self.worker_terminate_event,
            monitoring_store=self.monitoring_store,
            flush_interval=flush_interval,
            flush_every_num_message=flush_batch_size,
        )
        atexit.register(self._shutdown)
        self.worker.start()

    def get_inference_dataset(
        self,
        model_fqn: str,
        start_time: datetime,
        end_time: datetime,
        actual_value_required: bool,
    ):
        model = self.mlflow_client.get_model_by_fqn(model_fqn)

        return self.monitoring_store.get_inference_dataset(
            model_id=str(model.id),
            start_time=start_time,
            end_time=end_time,
            actual_value_required=actual_value_required,
        )

    @staticmethod
    def generate_hash_from_data(
        features: typing.Dict, timestamp: typing.Optional[datetime] = None
    ):
        data_str = json.dumps(features, sort_keys=True)
        if timestamp:
            data_str += timestamp.isoformat()
        return uuid.uuid5(uuid.NAMESPACE_X500, data_str).hex

    @staticmethod
    def generate_random_id():
        return uuid.uuid4().hex

    def _generate_prediction_packet_from_data(
        self, model_version_fqn: str, prediction: Prediction
    ) -> PredictionPacket:

        return PredictionPacket(
            model_version_id=self.monitoring_store.enable_monitoring_for_version(
                model_version_fqn
            ).id,
            prediction=prediction,
        )

    def _generate_actual_packet_from_data(self, model_version_fqn: str, actual: Actual):
        return ActualPacket(
            model_version_id=self.monitoring_store.enable_monitoring_for_version(
                model_version_fqn
            ).id,
            actual=actual,
        )

    def _put_in_queue(self, inference_packet: BasePacket):
        try:
            self.task_queue.put_nowait(inference_packet)
        except queue.Full as ex:
            raise MlFoundryException(
                "task queue is full\n"
                f"current task queue length is {self.task_queue.maxsize}\n"
                "consider increasing the task queue length using "
                f"{MAX_QUEUE_SIZE_ENV_VAR} environment variable"
            ) from ex

    @_check_shutdown
    def log_prediction(self, model_version_fqn: str, prediction: Prediction):
        prediction_packet = self._generate_prediction_packet_from_data(
            model_version_fqn=model_version_fqn, prediction=prediction
        )
        self._put_in_queue(prediction_packet)

    @_check_shutdown
    def log_actual(self, model_version_fqn: str, actual: Actual):
        actual_packet = self._generate_actual_packet_from_data(
            model_version_fqn=model_version_fqn, actual=actual
        )
        self._put_in_queue(actual_packet)

    @_check_shutdown
    def log_predictions(
        self, model_version_fqn: str, predictions: typing.List[Prediction]
    ):
        for prediction in predictions:
            self.log_prediction(
                model_version_fqn=model_version_fqn, prediction=prediction
            )

    @_check_shutdown
    def log_actuals(self, model_version_fqn: str, actuals: typing.List[Actual]):
        for actual in actuals:
            self.log_actual(model_version_fqn=model_version_fqn, actual=actual)

    def _flush(self):
        logger.debug(
            f"flushing task queue, {self.task_queue.qsize()} items in the queue"
        )
        self.task_queue.join()
        logger.debug("task queue flushed")

    def _shutdown(self):
        if self._is_shutdown:
            return
        logger.debug("shutting down worker and client")
        self._is_shutdown = True
        # NOTE: We initialize the monitoring store at first in the constructor
        # The task_queue, worker is defined later.
        # There is a chance that monitoring  store initialization will throw error,
        # in that case, shutdown will be called (__del__) but self.task_queue would not have
        # been initialized yet.
        if hasattr(self, "task_queue"):
            self._flush()
        if hasattr(self, "worker_terminate_event"):
            logger.debug("setting worker termination event")
            self.worker_terminate_event.set()
        if hasattr(self, "worker"):
            logger.debug("waiting for worker to terminate")
            self.worker.join()

    def __del__(self):
        self._shutdown()



from mlfoundry.monitoring.store.repositories.rest_monitoring_store import (
    RestMonitoringStore,
)


def get_monitoring_store(host_uri: str):
    # For now we have only one repo
    return RestMonitoringStore(host_uri=host_uri)



from mlflow.exceptions import MlflowException

from mlfoundry import amplitude


class MlFoundryException(MlflowException):
    def __init__(self, message):
        self.message = message
        amplitude.track(
            amplitude.Event.EXCEPTION,
            event_properties={"exception_message": message},
        )



from __future__ import print_function

import abc
import logging
import os
import threading
from timeit import default_timer
from types import TracebackType
from typing import Optional, Tuple, Type, Union

ExceptionType = Union[
    Tuple[Type[BaseException], BaseException, TracebackType],
    Tuple[None, None, None],
]

logger = logging.getLogger(__name__)


class BackgroundJob(threading.Thread):
    """
    Base class for background jobs in a MlFoundryRun

    Attributes:
        name: name of the thread
        interval: time in seconds to sleep between calls to `self._loop` i.e. between iterations
        _wake_up_event: threading.Event that can be set to wake the thread up from sleep between iterations
        _interrupted: loop until this flag is set to True. initialised to False
    """

    name_prefix = "MLFUnnamedThread"

    def __init__(self, name: str, interval: float) -> None:
        """
        Args:
            name: name of the thread
            interval: time in seconds to sleep between calls to `self._loop` i.e. between iterations
        """
        super().__init__(name=name, daemon=True)
        self.interval = interval
        self._wake_up_event = threading.Event()
        self._interrupted = False

    def run(self) -> None:
        """
        Call `self._loop` every `interval` seconds till `self._interrupted` is set to True and call `self._finish`
        before returning
        """
        execution_time = self.interval  # to run immediately at start
        while not self._interrupted:
            if self.interval > 0 and not self._interrupted:
                timeout = max(0.0, self.interval - execution_time)
                self._wake_up_event.wait(timeout=timeout)
                self._wake_up_event.clear()
            start = default_timer()
            try:
                self._loop()
            except Exception as e:
                logger.warning(
                    f"Thread {self.name} encountered an internal error: {str(e)}"
                )
                # TODO (chiragjn): We should have a tolerance counter after which we should break out of the loop
            execution_time = default_timer() - start
        self._finish()

    @abc.abstractmethod
    def _loop(self):
        """
        Method body to call every `self.interval` seconds
        """
        raise NotImplementedError

    def _finish(self):
        """
        Method body to call before returning from Thread's run. Useful for cleaning up state.
        """
        pass

    def disable_sleep(self) -> None:
        """
        Disable sleeping between consecutive calls to `self._loop`
        """
        self.interval = 0

    def interrupt(self) -> None:
        """
        Set `self._interrupted` to True causing the run loop to break and trigger `self._finish`
        """
        self._interrupted = True
        self.wake_up()

    def wake_up(self) -> None:
        """
        Wake up the thread from its sleep between consecutive calls to `self._loop`
        """
        self._wake_up_event.set()

    def stop(self, disable_sleep: bool = False, timeout: Optional[int] = None) -> None:
        """
        Interrupt, wake up and wait for thread to finish by joining for `timeout` seconds

        Args:
            disable_sleep: if to disable sleep between consecutive calls to `self._loop`
            timeout: time in seconds to wait while joining the thread
        """
        if self.is_alive():
            if disable_sleep:
                self.disable_sleep()
            self.wake_up()
            self.interrupt()
            self.join(timeout)


def make_log(message: str) -> str:
    """
    Add process pid and thread name as prefix to `message`
    Args:
        message: message to prepend to

    Returns:
        str: message with process pid and thread name as prepended
    """
    # TODO (chiragjn): Maybe just configure a logging.Handler with a logging.Formatter?
    return f"pid={os.getpid()} thread={threading.currentThread().getName()} {message}"



import enum
import posixpath
import sys
from typing import Union
from uuid import uuid4

from mlflow.entities import RunLog
from pydantic import BaseModel

from mlfoundry.constants import RUN_LOGS_DIR
from mlfoundry.exceptions import MlFoundryException
from mlfoundry.log_types.pydantic_base import PydanticBase
from mlfoundry.log_types.utils import save_artifact_helper, validate_key_name

PLOT_LOG_DIR = posixpath.join(RUN_LOGS_DIR, "plots")

PlotObjType = Union[
    "matplotlib.figure.Figure",
    "plotly.graph_objects.Figure",
    "matplotlib.pyplot",
]


@enum.unique
class Format(enum.Enum):
    SVG = "SVG"
    HTML = "HTML"
    PNG = "PNG"


def _is_matplotlib_figure(fig) -> bool:
    if "matplotlib" not in sys.modules:
        return False
    import matplotlib

    return isinstance(fig, matplotlib.figure.Figure)


def _is_matplotlib_plt(plt) -> bool:
    if "matplotlib" not in sys.modules:
        return False
    return getattr(plt, "__name__", "") == "matplotlib.pyplot"


def _is_plotly_figure(fig) -> bool:
    if "plotly" not in sys.modules:
        return False

    import plotly

    return isinstance(fig, plotly.graph_objects.Figure)


def get_plot_file_name(key: str, step: int, format: Format) -> str:
    return f"{key}-{step}-{uuid4()}.{format.value.lower()}"


class PlotArtifact(BaseModel):
    artifact_path: str
    format: Format

    class Config:
        allow_mutation = False
        use_enum_values = True


def _save_matplotlib_figure(
    figure: "matplotlib.figure.Figure",
    run: "mlfoundry.MlFoundryRun",
    key: str,
    step: int,
) -> PlotArtifact:
    supported_formats = figure.canvas.get_supported_filetypes().keys()
    if "svg" in supported_formats:
        format_ = Format.SVG
    elif "png" in supported_formats:
        format_ = Format.PNG
    else:
        raise MlFoundryException(
            f"Could not save {key} {figure} matplotlib figure"
            "in either SVG or PNG format"
        )
    file_path = get_plot_file_name(key=key, step=step, format=format_)
    artifact_path = posixpath.join(PLOT_LOG_DIR, file_path)
    with save_artifact_helper(run, artifact_path=artifact_path) as local_path:
        figure.savefig(local_path)
    return PlotArtifact(artifact_path=artifact_path, format=format_)


def _save_matplotlib_plt(
    plt: "matplotlib.pyplot",
    run: "mlfoundry.MlFoundryRun",
    key: str,
    step: int,
) -> PlotArtifact:
    figure = plt.gcf()
    return _save_matplotlib_figure(figure=figure, run=run, key=key, step=step)


def _save_plotly_figure(
    figure: "plotly.graph_objects.Figure",
    run: "mlfoundry.MlFoundryRun",
    key: str,
    step: int,
) -> PlotArtifact:
    format_ = Format.HTML
    file_path = get_plot_file_name(key=key, step=step, format=format_)
    artifact_path = posixpath.join(PLOT_LOG_DIR, file_path)
    with save_artifact_helper(run, artifact_path=artifact_path) as local_path:
        figure.write_html(local_path, include_plotlyjs="cdn", auto_open=False)
    return PlotArtifact(artifact_path=artifact_path, format=format_)


class Plot:
    def __init__(self, plot_obj: PlotObjType):
        self._plot_obj = plot_obj

    def _save_plot(
        self, run: "mlfoundry.MlFoundryRun", key: str, step: int
    ) -> PlotArtifact:
        if _is_matplotlib_plt(self._plot_obj):
            return _save_matplotlib_plt(plt=self._plot_obj, run=run, key=key, step=step)

        if _is_matplotlib_figure(self._plot_obj):
            return _save_matplotlib_figure(
                figure=self._plot_obj, run=run, key=key, step=step
            )

        if _is_plotly_figure(self._plot_obj):
            return _save_plotly_figure(
                figure=self._plot_obj, run=run, key=key, step=step
            )

        raise MlFoundryException(
            f"Unknown type: {type(self._plot_obj)}"
            "Supported types are, matplotlib.figure.Figure, matplotlib.pyplot"
            " and plotly.graph_objects.Figure"
        )

    def to_run_log(
        self,
        run: "mlfoundry.MlFoundryRun",
        key: str,
        step: int = 0,
    ) -> RunLog:
        validate_key_name(key)

        return PlotRunLogType(
            value=self._save_plot(run=run, key=key, step=step)
        ).to_run_log(key=key, step=step)

    def save(
        self,
        run: "mlfoundry.MlFoundryRun",
        key: str,
        step: int = 0,
    ):
        run_log = self.to_run_log(
            run=run,
            key=key,
            step=step,
        )
        run.mlflow_client.insert_run_logs(run_uuid=run.run_id, run_logs=[run_log])


class PlotRunLogType(PydanticBase):
    value: PlotArtifact

    @staticmethod
    def get_log_type() -> str:
        return "plot"


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import numpy as np
    import plotly.express as px

    import mlfoundry

    client = mlfoundry.get_client("https://app.devtest.truefoundry.tech")
    # client = mlfoundry.get_client("http://localhost:5000")

    run = client.create_run(project_name="plot-test")

    df = px.data.iris()
    fig = px.scatter(
        df,
        x="sepal_width",
        y="sepal_length",
        color="species",
        size="petal_length",
        hover_data=["petal_width"],
    )
    Plot(fig).save(run, "foo")

    df = px.data.tips()
    fig = px.histogram(
        df,
        x="total_bill",
        y="tip",
        color="sex",
        marginal="rug",
        hover_data=df.columns,
    )
    Plot(fig).save(run, "foo", step=1)

    names = ["group_a", "group_b", "group_c"]
    values = [1, 10, 100]
    plt.figure(figsize=(9, 3))
    plt.subplot(131)
    plt.bar(names, values)
    plt.subplot(132)
    plt.scatter(names, values)
    plt.subplot(133)
    plt.plot(names, values)
    plt.suptitle("Categorical Plotting")

    Plot(plt).save(run, "bar")

    plt.clf()

    data = {
        "a": np.arange(50),
        "c": np.random.randint(0, 50, 50),
        "d": np.random.randn(50),
    }
    data["b"] = data["a"] + 10 * np.random.randn(50)
    data["d"] = np.abs(data["d"]) * 100
    plt.scatter("a", "b", c="c", s="d", data=data)
    plt.xlabel("entry a")
    plt.ylabel("entry b")
    Plot(plt).save(run, "bar", 2)

    plt.clf()

    ax = plt.subplot()
    t = np.arange(0.0, 5.0, 0.01)
    s = np.cos(2 * np.pi * t)
    (line,) = plt.plot(t, s, lw=2)

    plt.annotate(
        "local max",
        xy=(2, 1),
        xytext=(3, 1.5),
        arrowprops=dict(facecolor="black", shrink=0.05),
    )

    plt.ylim(-2, 2)
    Plot(plt.gcf()).save(run, "bar", 3)



import typing

import pandas as pd
import whylogs
from google.protobuf.json_format import MessageToDict

from mlfoundry.dataset.types import DatasetStats, Profiles
from mlfoundry.dataset.whylogs_types import DataFrameSummary


def profile_to_summary(
    profile: whylogs.DatasetProfile, data: pd.DataFrame
) -> DataFrameSummary:
    dict_profile = MessageToDict(profile.to_summary(), preserving_proto_field_name=True)
    summary = DataFrameSummary(**dict_profile)
    for col in summary.columns.keys():
        if summary.columns[col].number_summary is None:
            summary.columns[col].set_value_count_summary(data[col])
    return summary


def build_stats(
    features: pd.DataFrame,
    predictions: typing.Optional[pd.Series],
    actuals: typing.Optional[pd.Series],
) -> typing.Tuple[DatasetStats, Profiles]:
    features_profile = whylogs.DatasetProfile("features")
    features_profile.track_dataframe(features)
    features_summary = profile_to_summary(features_profile, features)

    predictions_profile = None
    predictions_summary = None
    if predictions is not None:
        predictions_df = pd.DataFrame()
        predictions_df["predictions"] = predictions
        predictions_profile = whylogs.DatasetProfile("predictions")
        predictions_profile.track_dataframe(predictions_df)
        predictions_summary = profile_to_summary(predictions_profile, predictions_df)

    actuals_profile = None
    actuals_summary = None
    if actuals is not None:
        actuals_df = pd.DataFrame()
        actuals_df["actuals"] = actuals
        actuals_profile = whylogs.DatasetProfile("actuals")
        actuals_profile.track_dataframe(actuals_df)
        actuals_summary = profile_to_summary(actuals_profile, actuals_df)

    dataset_stats = DatasetStats(
        features=features_summary,
        predictions=predictions_summary,
        actuals=actuals_summary,
    )

    profiles = Profiles(
        features=features_profile,
        actuals=actuals_profile,
        predictions=predictions_profile,
    )
    return dataset_stats, profiles



"""
Define all the env variable names that users
can use
"""

TRACKING_HOST = "MLF_HOST"
API_KEY = "MLF_API_KEY"

## NOTE: Global means this API key will be
# shared between servicefoundry and mlfoundry
TRACKING_HOST_GLOBAL = "TFY_HOST"
API_KEY_GLOBAL = "TFY_API_KEY"
##

TRACKING_TOKEN = "MLFLOW_TRACKING_TOKEN"



import logging
import sys

from mlflow.entities import Feature

from mlfoundry.enums import *
from mlfoundry.log_types import Image, Plot
from mlfoundry.log_types.artifacts.model import CustomMetric, ModelSchema, ModelVersion
from mlfoundry.login import login
from mlfoundry.mlfoundry_api import get_client
from mlfoundry.mlfoundry_run import MlFoundryRun
from mlfoundry.monitoring.entities import Actual, Prediction, PredictionData
from mlfoundry.schema import Schema
from mlfoundry.version import __version__

__all__ = [
    "FileFormat",
    "ModelFramework",
    "DataSlice",
    "ModelType",
    "Schema",
    "get_client",
    "__version__",
    "MlFoundryRun",
    "login",
    "Image",
    "Plot",
    "Prediction",
    "PredictionData",
    "Actual",
    "Feature",
    "ModelSchema",
    "ModelVersion",
    "CustomMetric",
]


def _configure_logger(level):
    # TODO (chiragjn): This is just a temporary way to configure logging.
    #                  Need to move it in client or run init with option to set verbosity
    logger = logging.getLogger(__name__)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(level)
    handler.setFormatter(
        logging.Formatter(
            "[mlfoundry] %(asctime)s %(levelname)s %(message)s",
            datefmt="%Y-%m-%dT%H:%M:%S%z",
        )
    )
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)
    logger.propagate = False


_configure_logger(logging.INFO)



from datetime import datetime
from typing import Dict, List, Optional

from pydantic import BaseModel

from mlfoundry.monitoring.entities import DatasetData, EntityValue


class ClassPrediction(BaseModel):
    label: str
    score: float


class MlModelPrediction(BaseModel):
    value: EntityValue
    probabilities: Optional[List[ClassPrediction]]
    shap_values: Optional[Dict[str, float]]
    occurred_at: datetime

    class Config:
        smart_union = True


class Data(BaseModel):
    data_id: str
    features: Dict[str, EntityValue]
    prediction: MlModelPrediction
    raw_data: Dict

    class Config:
        smart_union = True


class ActualData(BaseModel):
    data_id: str
    actual: EntityValue

    class Config:
        smart_union = True


class BatchInsertRequest(BaseModel):
    model_version_id: str
    tag: Optional[str]
    items: List[Data]

    def to_json_dict(self):
        request_dict = self.dict()
        for item in request_dict["items"]:
            item["prediction"]["occurred_at"] = item["prediction"][
                "occurred_at"
            ].isoformat()
        return request_dict


class BatchUpdateActualRequest(BaseModel):
    model_version_id: str
    items: List[ActualData]


class GetDatasetResponse(BaseModel):
    total_rows: int
    data: List[DatasetData]



import argparse
import enum
import json
import logging
import os
import tempfile
from typing import Callable, Dict, List, Mapping, Optional, Sequence, Union
from weakref import ReferenceType

from mlflow.entities import RunStatus
from pytorch_lightning.callbacks.model_checkpoint import ModelCheckpoint
from pytorch_lightning.loggers.base import LightningLoggerBase, rank_zero_experiment
from pytorch_lightning.utilities.distributed import rank_zero_only
from pytorch_lightning.utilities.logger import (
    _add_prefix,
    _convert_params,
    _sanitize_callable_params,
)

import mlfoundry as mlf
from mlfoundry.enums import EnumMissingMixin
from mlfoundry.mlfoundry_run import MlFoundryRun

logger = logging.getLogger(__name__)

LIGHTNING_ARTIFACTS_PATH = "lightning/"
LIGHTNING_MODELS_PATH = os.path.join(LIGHTNING_ARTIFACTS_PATH, "models")


class LogModelStrategy(EnumMissingMixin, enum.Enum):
    # To keep it consistent with transformers
    NO = "no"
    BEST_ONLY = "best"
    BEST_PLUS_LATEST = "best_plus_latest"
    CHECKPOINTS_ON_TRAIN_END = "checkpoints_on_train_end"
    # TODO: Add ALL strategy, and how to handle that?


def _get_artifact_path_for_checkpoint(checkpoint_path):
    return os.path.join(LIGHTNING_MODELS_PATH, os.path.basename(checkpoint_path))


class MlFoundryLightningLogger(LightningLoggerBase):
    """
    Pytorch Lightning Logger for tracking training run on MLFoundry

    Examples:

        ```
        import mlfoundry as mlf
        mlf.login()
        # or set API key via `MLF_API_KEY` env variable

        from pytorch_lightning import Trainer
        from pytorch_lightning.callbacks import ModelCheckpoint
        from mlfoundry.integrations.lightning import MlFoundryLightningLogger, LogModelStrategy

        mlf_logger = MlFoundryLightningLogger(
            project_name="lightning",
            run_name="my-lightning-run",
            flatten_params=True,
            log_model_strategy=LogModelStrategy.BEST_PLUS_LATEST,
        )

        checkpoint_callback = ModelCheckpoint(
            monitor='val_accuracy',
            mode='max',
            save_last=True,
            save_top_k=-1
            )
        trainer = Trainer(
            logger=mlf_logger,
            callbacks=[checkpoint_callback], # our model checkpoint callback
            ..
            )
        trainer.fit(model, training_loader, validation_loader)
        ```

        Logger can also be created from an existing run

        ```
        import mlfoundry as mlf
        mlf.login()
        # or set API key via `MLF_API_KEY` env variable

        from pytorch_lightning import Trainer
        from pytorch_lightning.callbacks import ModelCheckpoint
        from mlfoundry.integrations.lightning import MlFoundryLightningLogger, LogModelStrategy

        client = mlf.get_client()
        run = client.create_run(project_name="lightning")

        mlf_logger = MlFoundryLightningLogger.from_run(
            run=run,
            auto_end_run=False,
            flatten_params=True,
            log_model_strategy=LogModelStrategy.BEST_PLUS_LATEST,
        )

        checkpoint_callback = ModelCheckpoint(
            monitor='val_accuracy',
            mode='max',
            save_last=True,
            save_top_k=-1
            )
        trainer = Trainer(
            logger=mlf_logger,
            callbacks=[checkpoint_callback], # our model checkpoint callback
            ..
            )
        trainer.fit(model, training_loader, validation_loader)

        run.end()
        ```
    """

    def __init__(
        self,
        project_name: str,
        run_name: Optional[str] = None,
        log_model_strategy: Union[str, LogModelStrategy] = LogModelStrategy.NO,
        flatten_params: Optional[bool] = False,
        prefix: str = "",
        agg_key_funcs: Optional[
            Mapping[str, Callable[[Sequence[float]], float]]
        ] = None,
        agg_default_func: Optional[Callable[[Sequence[float]], float]] = None,
        **kwargs,
    ):
        """
        Args:
            project_name (str): name of the project to create the run under
            run_name (Optional[str], optional): name of the run. When not provided a run name is automatically generated
            log_model_strategy (LogModelStrategy, optional): The strategy to use for logging models
                - LogModelStrategy.NO (default): Do not log any models
                - LogModelStrategy.BEST_ONLY: Log only the best model checkpoint.
                - LogModelStrategy.BEST_PLUS_LATEST: Log both the latest checkpoint and the best checkpoint (if
                    available) and different from the latest checkpoint. Note that `save_last` argument must be set
                    in the `ModelCheckpoint` callback, else latest checkpoints wont be saved/logged.
                - LogModelStrategy.CHECKPOINTS_ON_TRAIN_END: Log all available checkpoints that are available at the
                    end of training. Note that `save_top_k` must be set in the `ModelCheckpoint` callback,
                    else models from the previous steps wont be saved/logged

                In addition to checkpoints a metadata.json file is logged which stores references to the best checkpoint
                and the latest checkpoint and their corresponding scores

                The files will be logged under `mlf/lightning/models/` under the run artifacts.
            flatten_params (bool, optional): Flatten hierarchical dict
                By default, this is `False`

                For e.g.
                when set to `True`:
                `{'a': {'b': 'c'}}` will be logged as -> `{'a/b': 'c'}`

                when set to `False`:
                Individual keys will be logged ie) `'a':{'b':'c'}`
            prefix (str, optional): This prefix will be added in front of all metric_name logged.

        """
        super().__init__(agg_key_funcs=agg_key_funcs, agg_default_func=agg_default_func)

        log_model_strategy = LogModelStrategy(log_model_strategy)

        self._project_name = project_name
        self._run_name = run_name
        self._log_model_strategy = log_model_strategy
        self._flatten_params = flatten_params
        self._auto_end_run = True
        self._prefix = prefix

        self._run = None
        self._checkpoint_callback = None
        self._available_model_paths: List[str] = []

    @property
    @rank_zero_experiment  # Returns the real experiment on rank 0 and otherwise the DummyExperiment
    def run(self) -> MlFoundryRun:
        if self._run:
            return self._run

        if not self._run:
            self._run = mlf.get_client().create_run(
                project_name=self._project_name,
                run_name=self._run_name,
                log_system_metrics=True,
            )
            self._run_name = self._run.run_name
            self._auto_end_run = True

        return self._run

    @property
    def name(self) -> str:
        return self.run.project_name

    @property
    def version(self) -> str:
        return self.run.run_name

    @classmethod
    def from_run(
        cls,
        run: MlFoundryRun,
        log_model_strategy: Union[str, LogModelStrategy] = LogModelStrategy.NO,
        flatten_params: Optional[bool] = False,
        auto_end_run: Optional[bool] = False,
        **kwargs,
    ):
        """
        Create a MLFoundry Lightning Logger from an existing MLFoundry run instance.

        Args:
            run (MlFoundryRun): `MlFoundry` run instance
            log_model_strategy (LogModelStrategy, optional): The strategy to use for logging models
                - LogModelStrategy.NO (default): Do not log any models
                - LogModelStrategy.BEST_ONLY: Log only the best model checkpoint.
                - LogModelStrategy.BEST_PLUS_LATEST: Log both the latest checkpoint and the best checkpoint (if
                    available) and different from the latest checkpoint. Note that `save_last` argument must be set
                    in the `ModelCheckpoint` callback, else latest checkpoints wont be saved/logged.
                - LogModelStrategy.CHECKPOINTS_ON_TRAIN_END: Log all available checkpoints that are available at the
                    end of training. Note that `save_top_k` must be set in the `ModelCheckpoint` callback,
                    else models from the previous steps wont be saved/logged

                In addition to checkpoints a metadata.json file is logged which stores references to the best checkpoint
                and the latest checkpoint and their corresponding scores

                The files will be logged under `mlf/lightning/models/` under the run artifacts.
            flatten_params (bool, optional): Flatten hierarchical dict
                By default, this is `False`

                For e.g.
                when set to `True`:
                `{'a': {'b': 'c'}}` will be logged as -> `{'a/b': 'c'}`

                when set to `False`:
                Individual keys will be logged ie) `'a':{'b':'c'}`
            auto_end_run (bool, optional): if to end the run when training finishes. By default, this is `False`
        """
        instance = cls(
            project_name=run.project_name,
            run_name=run.run_name,
            log_model_strategy=log_model_strategy,
            flatten_params=flatten_params,
            **kwargs,
        )
        instance._run = run
        instance._auto_end_run = auto_end_run

        return instance

    @rank_zero_only  # runs only on global rank 0. To prevent multiple logging if there is distributed training
    def log_hyperparams(self, params: argparse.Namespace, *args, **kwargs) -> None:
        """Record hyperparameters.

        Args:
            params (argparse.Namespace): containing the hyperparameters
            args: Optional positional arguments, depends on the specific logger being used
            kwargs: Optional keyword arguments, depends on the specific logger being used
        """
        params = _convert_params(params)  # convert to dict
        params = _sanitize_callable_params(params)  # convert callable params

        self.run.log_params(param_dict=params, flatten_params=self._flatten_params)

    @rank_zero_only
    def log_metrics(
        self, metrics: Dict[str, float], step: Optional[int] = None
    ) -> None:
        """
        Records metrics.
        This method logs metrics as as soon as it received them. If you want to aggregate
        metrics for one specific `step`, use the
        :meth:`~pytorch_lightning.loggers.base.Logger.agg_and_log_metrics` method.

        Args:
            metrics: Dictionary with metric names as keys and measured quantities as values
            step: Step number at which the metrics should be recorded
        """
        metrics = _add_prefix(metrics, self._prefix, "/")
        self.run.log_metrics(metric_dict=metrics, step=step if step else 0)

    def _log_model_on_train_end(
        self, checkpoint_callback: "ReferenceType[ModelCheckpoint]", temp_dir
    ):

        latest_checkpoint_path = checkpoint_callback.last_model_path
        best_checkpoint_path = checkpoint_callback.best_model_path
        metadata = {"checkpoints": {}}

        checkpoints_to_log: List[str] = []
        if self._log_model_strategy == LogModelStrategy.BEST_ONLY:
            latest_checkpoint_path = None
            if best_checkpoint_path is not None and os.path.exists(
                best_checkpoint_path
            ):
                checkpoints_to_log.append(best_checkpoint_path)
            else:
                logger.warning(
                    f"`log_model_strategy` was set to BEST_ONLY but CheckpointCallback did not save any best "
                    f"checkpoint. Cannot automatically log model"
                )

        elif self._log_model_strategy == LogModelStrategy.CHECKPOINTS_ON_TRAIN_END:
            for model_path in self._available_model_paths:
                if os.path.exists(model_path):
                    checkpoints_to_log.append(model_path)

        elif self._log_model_strategy == LogModelStrategy.BEST_PLUS_LATEST:
            if latest_checkpoint_path and os.path.exists(latest_checkpoint_path):
                checkpoints_to_log.append(latest_checkpoint_path)
            if (
                best_checkpoint_path
                and best_checkpoint_path != latest_checkpoint_path
                and os.path.exists(best_checkpoint_path)
            ):
                checkpoints_to_log.append(best_checkpoint_path)
        else:
            raise NotImplementedError("Unreachable!")

        for checkpoint_path in checkpoints_to_log:
            checkpoint_name = os.path.basename(checkpoint_path)
            artifact_path = os.path.join(LIGHTNING_MODELS_PATH, checkpoint_name)
            logger.debug(
                f"Logging checkpoint at {checkpoint_path!r} to {artifact_path!r}"
            )
            self._run.log_artifact(
                local_path=checkpoint_path, artifact_path=artifact_path
            )

        if best_checkpoint_path:
            best_model_artifact_path = _get_artifact_path_for_checkpoint(
                os.path.basename(best_checkpoint_path)
            )
            metadata["checkpoints"]["best"] = best_model_artifact_path

        if latest_checkpoint_path:
            latest_model_artifact_path = _get_artifact_path_for_checkpoint(
                os.path.basename(latest_checkpoint_path)
            )
            metadata["checkpoints"]["latest"] = latest_model_artifact_path

        metadata_file_name = "metadata.json"
        metadata_path = os.path.join(temp_dir, metadata_file_name)
        with open(metadata_path, "w") as f:
            json.dump(metadata, f)
        self._run.log_artifact(
            local_path=metadata_path,
            artifact_path=LIGHTNING_MODELS_PATH,
        )

    def after_save_checkpoint(
        self, checkpoint_callback: "ReferenceType[ModelCheckpoint]"
    ) -> None:
        """Event called after model checkpoint callback saves a new checkpoint.

        Args:
            checkpoint_callback: the model checkpoint callback instance
        """
        if self._log_model_strategy != LogModelStrategy.NO:
            self._checkpoint_callback = checkpoint_callback

            checkpoints = [
                checkpoint_callback.last_model_path,
                checkpoint_callback.best_model_path,
            ] + list(checkpoint_callback.best_k_models.keys())

            for model_path in checkpoints:
                if model_path not in self._available_model_paths:
                    self._available_model_paths.append(model_path)

    def finalize(self, status: str) -> None:
        """Event called at the end of training. Do any processing that is necessary to finalize an run.

        Args:
            status: Status that the run finished with (e.g. success, failed, aborted)
        """
        if self._checkpoint_callback:
            with tempfile.TemporaryDirectory() as temp_dir:
                self._log_model_on_train_end(self._checkpoint_callback, temp_dir)

        if self._auto_end_run and self.run:
            if status == "success":
                self.run.end()
            else:
                self.run.end(status=RunStatus.FAILED)

    def __del__(self):
        if self._auto_end_run and self.run:
            self.run.end()



import logging
import os
import typing

import jwt

from mlfoundry import env_vars
from mlfoundry.exceptions import MlFoundryException
from mlfoundry.login import get_stored_api_key, login
from mlfoundry.tracking.auth_service import AuthService
from mlfoundry.tracking.entities.user_info import UserInfo

logger = logging.getLogger(__name__)


def is_notebook_or_interactive():
    try:
        shell = get_ipython().__class__
        if shell.__name__ == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        if shell.__name__ == "TerminalInteractiveShell":
            return True  # Terminal running IPython
        if "google.colab" in str(shell):
            return True  # google colab notebook
        return False  # Other type (?)
    except NameError:
        return False


class Session:
    def __init__(self, auth_service: AuthService, tracking_uri: str):
        self.auth_service: AuthService = auth_service
        self.tracking_uri = tracking_uri

    def init_session(
        self,
        api_key: typing.Optional[str] = None,
    ) -> UserInfo:
        final_api_key = (
            api_key
            or os.getenv(env_vars.API_KEY)
            or os.getenv(env_vars.API_KEY_GLOBAL)
            or get_stored_api_key(self.tracking_uri)
        )
        if final_api_key is not None:
            token = self.auth_service.get_token(api_key=final_api_key)
            os.environ[env_vars.TRACKING_TOKEN] = token
            return Session._get_user_info(token)
        # if API key is not present,
        # then take a look if MLFLOW_TRACKING_TOKEN itself has been set.
        # this will be used in sfy.
        existing_token = os.getenv(env_vars.TRACKING_TOKEN, "")
        if existing_token:
            logger.info("API key is not present. Using existing tracking token")
            return Session._get_user_info(existing_token)

        if is_notebook_or_interactive() and login(self.tracking_uri):
            return self.init_session(get_stored_api_key(self.tracking_uri))

        raise MlFoundryException(
            "Could not find API key. Please use `mlfoundry login` to log in."
        )

    @staticmethod
    def _get_user_info(jwt_token: str) -> UserInfo:
        # TODO:- should probably do this whole thing in the backend.
        try:
            json_body = jwt.decode(
                jwt_token,
                options={
                    "verify_signature": False,
                    "verify_aud": False,
                },
            )
            user_info = UserInfo(
                user_id=json_body["username"],
                email=json_body["email"],
            )
            logger.info("Welcome user! You are logged in as %s", user_info.user_id)
        except Exception as ex:
            logger.exception("failed to get user info, %s", ex)
            raise ex

        return user_info



import typing

from pydantic import BaseModel, Field


class RocCurve(BaseModel):
    fpr: typing.List[float]
    tpr: typing.List[float]
    thresholds: typing.List[float]


class PrCurve(BaseModel):
    precision: typing.List[float]
    recall: typing.List[float]
    thresholds: typing.List[float]


class ConfusionMatrix(BaseModel):
    class_names: typing.List[str]
    matrix: typing.List[typing.List[float]]


class ClassificationReport(BaseModel):
    precision: float
    recall: float
    f1_score: float = Field(None, alias="f1-score")
    support: float



import argparse
import importlib
import json
import os
import tempfile
import time
import typing
from collections.abc import Mapping
from urllib.parse import urljoin, urlsplit

import mlflow
import numpy as np

from mlfoundry import constants, env_vars
from mlfoundry.exceptions import MlFoundryException


def get_module(
    module_name: str, error_message: typing.Optional[str] = None, required: bool = False
):
    try:
        return importlib.import_module(module_name)
    except Exception as ex:
        msg = error_message or f"Error importing module {module_name}"
        if required:
            raise MlFoundryException(msg) from ex


def resolve_tracking_uri(tracking_uri: typing.Optional[str]):
    return (
        tracking_uri
        or os.getenv(env_vars.TRACKING_HOST)
        or os.getenv(env_vars.TRACKING_HOST_GLOBAL)
        or constants.DEFAULT_TRACKING_URI
    )


def append_path_to_rest_tracking_uri(tracking_uri: str):
    if urlsplit(tracking_uri).netloc.startswith("localhost"):
        return tracking_uri
    return urljoin(tracking_uri, "/api/ml")


def append_monitoring_path_to_rest_tracking_uri(tracking_uri: str):
    if urlsplit(tracking_uri).netloc.startswith("localhost"):
        return tracking_uri
    return urljoin(tracking_uri, "/api/monitoring")


class NumpyEncoder(json.JSONEncoder):
    """Special json encoder for numpy types"""

    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


ParamsType = typing.Union[typing.Mapping[str, typing.Any], argparse.Namespace]


def process_params(params: ParamsType) -> typing.Dict[str, typing.Any]:
    if isinstance(params, Mapping):
        return params
    if isinstance(params, argparse.Namespace):
        return vars(params)
    # TODO: add absl support if required
    # move to a different file then
    raise MlFoundryException(
        "params should be either argparse.Namespace or a Mapping (dict) type"
    )


def log_artifact_blob(
    mlflow_client: mlflow.tracking.MlflowClient,
    run_id: str,
    blob: typing.Union[str, bytes],
    file_name: str,
    artifact_path: typing.Optional[str] = None,
):
    with tempfile.TemporaryDirectory(prefix=run_id) as tmpdirname:
        local_path = os.path.join(tmpdirname, file_name)
        mode = "wb" if isinstance(blob, bytes) else "w"
        with open(local_path, mode) as local_file:
            local_file.write(blob)
        mlflow_client.log_artifact(run_id, local_path, artifact_path=artifact_path)


def mapping_to_mlflow_metric(
    metrics: typing.Mapping[str, float],
    timestamp: typing.Optional[int] = None,
    step: int = 0,
) -> typing.Dict[str, mlflow.entities.Metric]:
    if timestamp is None:
        timestamp = int(time.time() * 1000)
    mlflow_metrics = {
        key: mlflow.entities.Metric(key, value, timestamp, step=step)
        for key, value in metrics.items()
    }
    return mlflow_metrics


def flatten_dict(
    input_dict: typing.Dict[typing.Any, typing.Any], parent_key="", sep="."
) -> typing.Dict[str, typing.Any]:
    """Flatten hierarchical dict, e.g. ``{'a': {'b': 'c'}} -> {'a.b': 'c'}``.
    All the keys will be converted to str.

    Args:
        input_dict: Dictionary containing the keys
        sep: Delimiter to express the hierarchy. Defaults to ``'.'``.

    Returns:
        Flattened dict.

    Examples:
        >>> flatten_dict({'a': {'b': 'c'}})
        {'a.b': 'c'}
        >>> flatten_dict({'a': {'b': 123}})
        {'a.b': 123}
        >>> flatten_dict({'a': {'b': 'c'}}, parent_key="param")
        {'param.a.b': 'c'}
    """
    new_dict_items = []
    for k, v in input_dict.items():
        new_key = parent_key + sep + str(k) if parent_key else k
        if isinstance(v, Mapping):
            new_dict_items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            new_dict_items.append((new_key, v))
    return dict(new_dict_items)
